local pcall   = pcall
local pairs   = pairs
local ipairs  = ipairs
local rawset  = rawset
local require = require

local table = require "table"
local loop  = require "loop"
local oo    = require "loop.scoped"

module("luaunit.framework.TestResult",
	loop.define(oo.class({}, require "luaunit.framework.Constructable")),
	loop.seeapi(oo) )

local TestFailure      = require "luaunit.framework.TestFailure"
local FailureException = require "luaunit.framework.FailureException"

protected = {
	fRunTests = 0,
	fStop = false,
}

function constructor(self)
	rawset(protected(self), "fErrors", {})
	rawset(protected(self), "fFailures", {})
	rawset(protected(self), "fListeners", {})
end

function public:addError(test, error)
	local test_error = TestFailure(test, error)
	table.insert(self.fErrors, test_error)
	for listener in pairs(self.fListeners) do
		listener:addError(test, error)
	end
end

function public:addFailure(test, failure)
	local test_failure = TestFailure(test, failure)
	table.insert(self.fFailures, test_failure)
	for listener in pairs(self.fListeners) do
		listener:addFailure(test, failure)
	end
end

function public:addListener(listener)
	self.fListeners[listener] = true
end

function public:endTest(test)
	for listener in pairs(self.fListeners) do
		listener:endTest(test)
	end
end

function public:errorCount()
	return table.getn(self.fErrors)
end

function public:errors()
	local errors = {}
	for _, error in ipairs(self.fErrors) do
		table.insert(errors, error)
	end
	return errors
end

function public:failureCount()
	return table.getn(self.fFailures)
end

function public:failures()
	local failures = {}
	for _, error in ipairs(self.fFailures) do
		table.insert(failures, error)
	end
	return failures
end

function public:removeListener(listener)
	self.fListeners[listener] = nil
end

function public:run(test)
	self:startTest(test)
	local success, message = pcall(test.runBare, test)
	if not success then
		if instanceof(message, FailureException)
			then self:addFailure(test, message)
			else self:addError(test, message)
		end
	end
	self:endTest(test)
end

function public:runCount()
	return self.fRunTests
end

function public:shouldStop()
	return self.fStop
end

function public:startTest(test)
	self.fRunTests = self.fRunTests + test:countTestCases()
	for listener in pairs(self.fListeners) do
		listener:startTest(test)
	end
end

function public:stop()
	self.fStop = true
end

function public:wasSuccessful()
	return (self:errorCount() == 0) and (self:failureCount() == 0)
end
