require("luaunit.framework.Assert")
require("luaunit.utils.ParameterGenerator")

local tests

-------------------------------------------------------------------------------

local function tabletostring(...)
	for index = 1, table.getn(arg) do
		arg[index] = tostring(arg[index])
	end
	return table.concat(arg, ", ")
end

local function exerciseAssert(object, method, paramList, expected, test_message)
	local message = "message to be returned"
	for _, params in ipairs(paramList) do
		if test_message then
			params[test_message] = message
		end
		local actual, error_message = pcall(object[method], object, params[1], params[2], params[3], params[4], params[5], params[6], params[7], params[8], params[9])
		assert(actual == expected, "'" .. method .. "' failed for parameters (" .. tabletostring(params[1], params[2], params[3], params[4], params[5], params[6], params[7], params[8], params[9]) .. ")")
		if test_message then
			if expected
				then assert(error_message == nil, "'" .. method .. "' returned a value for parameters (" .. tabletostring(params[1], params[2], params[3], params[4], params[5], params[6], params[7], params[8], params[9]) .. ")")
				else assert(string.find(error_message.message, message) == 1, "'" .. method .. "' not returned the message for parameters (" .. tabletostring(params[1], params[2], params[3], params[4], params[5], params[6], params[7], params[8], params[9]) .. ")")
			end
			params[test_message] = nil
		end
	end
end

local function getEqualParams()

	local func = function() end
	local thread = coroutine.create(function() end)
	local tab = {}

	local equalParams = {
		{},
		{ true, true },
		{ false, false },
		{ -1, -1 },
		{ 0, 0 },
		{ 1, 1 },
		{ "string", "string" },
		{ func, func },
		{ thread, thread },
		{ tab, tab },
	}

	return equalParams
end

local function getDifferentParams()
	local differentParams = {
		{ true, false },
		{ false, true },
		{ -1.1, 1.1 },
		{ 0, 1.1 },
		{ -1.1, 0 },
		{ 1.1, -1.1 },
		{ "string", "strong" },
		{ function() end, function() end },
		{ coroutine.create(function() end), coroutine.create(function() end) },
		{ {n=2}, {n=2} },
	}
	for type_pair in tests.generator:parameters{"type_name", "type_name"} do
		if type_pair[1] ~= type_pair[2] then
			for params in tests.generator:parameters(type_pair) do
				table.insert(differentParams, { params[1], params[2], n=2 })
			end
		end
	end
	
	return differentParams
end

local function getAllValues()
	local allValues = {}
	for type_pair in tests.generator:parameters{"type_name"} do
		for params in tests.generator:parameters(type_pair) do
			table.insert(allValues, params)
		end
	end

	return allValues
end

local function removeValue(paramList, value)
	local pos
	for index, params in ipairs(paramList) do
		if params[1] == value then pos = index end
	end
	table.remove(paramList, pos)
end

-------------------------------------------------------------------------------

tests = {
	generator = luaunit.utils.ParameterGenerator()
}
tests.generator:defineValueSet("type_name", { "nil", "boolean", "number", "string", "function", "thread", "table" } )

function tests:setUp()
	self.assert = luaunit.framework.Assert()
end

function tests:tearDown()
	self.assert = nil
end

function tests:Equality(test_name)
	local equal = getEqualParams()
	local diff = getDifferentParams()
	exerciseAssert(self.assert, "assertEquals", equal, true)
	exerciseAssert(self.assert, "assertEquals", equal, true, 3)
	exerciseAssert(self.assert, "assertEquals", diff, false)
	exerciseAssert(self.assert, "assertEquals", diff, false, 3)
	exerciseAssert(self.assert, "assertSame", equal, true)
	exerciseAssert(self.assert, "assertSame", equal, true, 3)
	exerciseAssert(self.assert, "assertSame", diff, false)
	exerciseAssert(self.assert, "assertSame", diff, false, 3)
	exerciseAssert(self.assert, "assertNotSame", diff, true)
	exerciseAssert(self.assert, "assertNotSame", diff, true, 3)
	exerciseAssert(self.assert, "assertNotSame", equal, false)
	exerciseAssert(self.assert, "assertNotSame", equal, false, 3)
end

function tests:NilReferences(test_name)
	local notnil = getAllValues()
	removeValue(notnil, nil)
	exerciseAssert(self.assert, "assertNil", { {n=2} }, true)
	exerciseAssert(self.assert, "assertNil", { {n=2} }, true, 2)
	exerciseAssert(self.assert, "assertNil", notnil, false)
	exerciseAssert(self.assert, "assertNil", notnil, false, 2)
	exerciseAssert(self.assert, "assertNotNil", notnil, true)
	exerciseAssert(self.assert, "assertNotNil", notnil, true, 2)
	exerciseAssert(self.assert, "assertNotNil", { {n=2} }, false)
	exerciseAssert(self.assert, "assertNotNil", { {n=2} }, false, 2)
end

function tests:Conditionals(test_name)
	local falseValues = { {}, { false } }
	local trueValues = getAllValues()
	removeValue(trueValues, nil)
	removeValue(trueValues, false)
	exerciseAssert(self.assert, "assertTrue", trueValues, true)
	exerciseAssert(self.assert, "assertTrue", trueValues, true, 2)
	exerciseAssert(self.assert, "assertTrue", falseValues, false)
	exerciseAssert(self.assert, "assertTrue", falseValues, false, 2)
	exerciseAssert(self.assert, "assertFalse", falseValues, true)
	exerciseAssert(self.assert, "assertFalse", falseValues, true, 2)
	exerciseAssert(self.assert, "assertFalse", trueValues, false)
	exerciseAssert(self.assert, "assertFalse", trueValues, false, 2)
end

return tests