require "loop.base"
require "luaunit.framework.TestResult"

local tests

-------------------------------------------------------------------------------

local DefaultTestListener = loop.base.class {
	starts = 0,
	ends = 0,
	failures = 0,
	errors = 0,
	addError = function(self, test, error)
		self.errors = self.errors + 1
	end,
	addFailure = function(self, test, failure)
		self.failures = self.failures + 1
	end,
	endTest = function(self, test)
		self.ends = self.ends + 1
	end,
	startTest = function(self, test)
		self.starts = self.starts + 1
	end,
}

local DefaultTest = loop.base.class {
	runs = 0,
	countTestCases = function() return 1 end,
	runBare = function(self, test, error)
		self.runs = self.runs + 1
		if self.runTest then self:runTest() end
	end,
}

local function unexpectedError()
	error(tests.error)
end
local function assertionFailure()
	error(tests.failure)
end

-------------------------------------------------------------------------------

local success, failure = pcall(luaunit.framework.Assert.fail, luaunit.framework.Assert)
tests = {
	error = "file.lua:69: Error!",
	failure = failure,
}
function tests:setUp()
	self.test = DefaultTest()
	self.failure_test = DefaultTest{ runTest = assertionFailure }
	self.error_test = DefaultTest{ runTest = unexpectedError }
	self.listener = DefaultTestListener()
	self.removed_listener = DefaultTestListener()
	self.result = luaunit.framework.TestResult()
end

function tests:tearDown()
	self.test = nil
	self.failure_test = nil
	self.error_test = nil
	self.listener = nil
	self.removed_listener = nil
	self.result = nil
end

function tests:Initialization(test_name)
	assert(self.result:errorCount() == 0, "'errorCount' is not 0 at '" .. test_name .. "'")
	assert(table.getn(self.result:errors()) == 0, "'errors' is not 0 at '" .. test_name .. "'")
	assert(self.result:failureCount() == 0, "'failureCount' is not 0 at '" .. test_name .. "'")
	assert(table.getn(self.result:failures()) == 0, "'failures' is not 0 at '" .. test_name .. "'")
	assert(self.result:runCount() == 0, "'runCount' is not 0 at '" .. test_name .. "'")
	assert(not self.result:shouldStop(), "'shouldStop' is not false at '" .. test_name .. "'")
	assert(self.result:wasSuccessful(), "'wasSuccessful' is not true at '" .. test_name .. "'")
end

function tests:ErrorAddition(test_name)
	self.result:addError(self.error_test, self.error)
	assert(self.result:errorCount() == 1, "'errorCount' is not 1 at '" .. test_name .. "'")
	local errors = self.result:errors()
	assert(errors[1]:errorMessage() == self.error, "'errors' failed at '" .. test_name .. "'")

	self.result:addError(self.error_test, self.error)
	assert(self.result:errorCount() == 2, "'errorCount' is not 2 at '" .. test_name .. "'")
	local errors = self.result:errors()
	assert(errors[1]:errorMessage() == self.error, "'errors' failed at '" .. test_name .. "'")
	assert(errors[2]:errorMessage() == self.error, "'errors' failed at '" .. test_name .. "'")

	self.result:addError(self.error_test, self.error)
	assert(self.result:errorCount() == 3, "'errorCount' is not 3 at '" .. test_name .. "'")
	local errors = self.result:errors()
	assert(errors[1]:errorMessage() == self.error, "'errors' failed at '" .. test_name .. "'")
	assert(errors[2]:errorMessage() == self.error, "'errors' failed at '" .. test_name .. "'")
	assert(errors[3]:errorMessage() == self.error, "'errors' failed at '" .. test_name .. "'")
end

function tests:FailureAddition(test_name)
	self.result:addFailure(self.failure_test, self.failure)
	assert(self.result:failureCount() == 1, "'failureCount' is not 1 at '" .. test_name .. "'")
	local failures = self.result:failures()
	assert(failures[1]:errorMessage() == self.failure, "'failures' failed at '" .. test_name .. "'")

	self.result:addFailure(self.failure_test, self.failure)
	assert(self.result:failureCount() == 2, "'failureCount' is not 2 at '" .. test_name .. "'")
	local failures = self.result:failures()
	assert(failures[1]:errorMessage() == self.failure, "'failures' failed at '" .. test_name .. "'")
	assert(failures[2]:errorMessage() == self.failure, "'failures' failed at '" .. test_name .. "'")

	self.result:addFailure(self.failure_test, self.failure)
	assert(self.result:failureCount() == 3, "'failureCount' is not 3 at '" .. test_name .. "'")
	local failures = self.result:failures()
	assert(failures[1]:errorMessage() == self.failure, "'failures' failed at '" .. test_name .. "'")
	assert(failures[2]:errorMessage() == self.failure, "'failures' failed at '" .. test_name .. "'")
	assert(failures[3]:errorMessage() == self.failure, "'failures' failed at '" .. test_name .. "'")
end

function tests:CorrectRun(test_name)
	self.result:run(self.test)
	assert(self.result:errorCount() == 0, "'errorCount' is not 0 at '" .. test_name .. "'")
	assert(table.getn(self.result:errors()) == 0, "'errors' is not 0 at '" .. test_name .. "'")
	assert(self.result:failureCount() == 0, "'failureCount' is not 0 at '" .. test_name .. "'")
	assert(table.getn(self.result:failures()) == 0, "'failures' is not 0 at '" .. test_name .. "'")
	assert(self.result:runCount() == 1, "'runCount' is not 1 at '" .. test_name .. "'")
	assert(not self.result:shouldStop(), "'shouldStop' is not false at '" .. test_name .. "'")
	assert(self.result:wasSuccessful(), "'wasSuccessful' is not true at '" .. test_name .. "'")
end

function tests:ErrorRun(test_name)
	self.result:run(self.error_test)
	assert(self.result:errorCount() == 1, "'errorCount' is not 1 at '" .. test_name .. "'")
	local errors = self.result:errors()
	assert(table.getn(errors) == 1, "'errors' is not 1 at '" .. test_name .. "'")
	assert(string.find(errors[1]:errorMessage(), self.error) ~= nil, "'errors' has not the expected error, has " .. tostring(errors[1]:errorMessage()) .. " instead. At '" .. test_name .. "'")
	assert(self.result:failureCount() == 0, "'failureCount' is not 0 at '" .. test_name .. "'")
	assert(table.getn(self.result:failures()) == 0, "'failures' is not 0 at '" .. test_name .. "'")
	assert(self.result:runCount() == 1, "'runCount' is not 1 at '" .. test_name .. "'")
	assert(not self.result:shouldStop(), "'shouldStop' is not false at '" .. test_name .. "'")
	assert(not self.result:wasSuccessful(), "'wasSuccessful' is not false at '" .. test_name .. "'")
end

function tests:FailureRun(test_name)
	self.result:run(self.failure_test)
	assert(self.result:errorCount() == 0, "'errorCount' is not 0 at '" .. test_name .. "'")
	assert(table.getn(self.result:errors()) == 0, "'errors' is not 0 at '" .. test_name .. "'")
	assert(self.result:failureCount() == 1, "'failureCount' is not 1 at '" .. test_name .. "'")
	local failures = self.result:failures()
	assert(table.getn(failures) == 1, "'failures' is not 1 at '" .. test_name .. "'")
	assert(failures[1]:errorMessage() == self.failure, "'failures' has not the expected failure, has " .. tostring(failures[1]:errorMessage()) .. " instead. At '" .. test_name .. "'")
	assert(self.result:runCount() == 1, "'runCount' is not 1 at '" .. test_name .. "'")
	assert(not self.result:shouldStop(), "'shouldStop' is not false at '" .. test_name .. "'")
	assert(not self.result:wasSuccessful(), "'wasSuccessful' is not false at '" .. test_name .. "'")
end

function tests:StopCheck(test_name)
	assert(not self.result:shouldStop(), "'shouldStop' is not false at '" .. test_name .. "'")
	self.result:stop()
	assert(self.result:shouldStop(), "'shouldStop' is not true at '" .. test_name .. "'")
end

function tests:Listeners(test_name)
	self.result:addListener(self.listener)
	self.result:run(self.failure_test)
	assert(self.listener.starts == 1, "'run' has not informed listener at '" .. test_name .. "'")
	assert(self.listener.ends == 1, "'run' has not informed listener at '" .. test_name .. "'")
	assert(self.listener.failures == 1, "'run' has not informed listener at '" .. test_name .. "'")
	assert(self.listener.errors == 0, "'run' has not informed listener at '" .. test_name .. "'")
	self.result:addListener(self.removed_listener)
	self.result:removeListener(self.removed_listener)
	self.result:run(self.error_test)
	assert(self.listener.starts == 2, "'run' has not informed listener at '" .. test_name .. "'")
	assert(self.listener.ends == 2, "'run' has not informed listener at '" .. test_name .. "'")
	assert(self.listener.failures == 1, "'run' has not informed listener at '" .. test_name .. "'")
	assert(self.listener.errors == 1, "'run' has not informed listener at '" .. test_name .. "'")
	assert(self.removed_listener.starts == 0, "'removeListener' has not removed listener at '" .. test_name .. "'")
	assert(self.removed_listener.ends == 0, "'removeListener' has not removed listener at '" .. test_name .. "'")
	assert(self.removed_listener.failures == 0, "'removeListener' has not removed listener at '" .. test_name .. "'")
	assert(self.removed_listener.errors == 0, "'removeListener' has not removed listener at '" .. test_name .. "'")
end

return tests
