module("oil.tests.cases.Marshal")

CreateInterface = [[
	IDLTypes = {
		null     = true,
		void     = true,
		short    = true,
		ushort   = true,
		long     = true,
		ulong    = true,
		float    = true,
		double   = true,
		boolean  = true,
		char     = true,
		octet    = true,
		string   = true,
		any      = true,
	
		Object = "IDL:omg.org/CORBA/Object:1.0",
		struct = { name = "MyStruct",
			{type = oil.idl.short  , name = "short_value"},
			{type = oil.idl.ushort , name = "ushort_value"},
			{type = oil.idl.long   , name = "long_value"},
			{type = oil.idl.ulong  , name = "ulong_value"},
			{type = oil.idl.float  , name = "float_value"},
			{type = oil.idl.double , name = "double_value"},
			{type = oil.idl.boolean, name = "boolean_value"},
			{type = oil.idl.octet  , name = "octet_value"},
			{type = oil.idl.string , name = "string_value"},
		},
		union = { name = "MyUnion",
			switch = oil.idl.ulong,
			options = {
				{label = 8, type = oil.idl.short  , name = "short_value"},
				{label = 7, type = oil.idl.ushort , name = "ushort_value"},
				{label = 6, type = oil.idl.long   , name = "long_value"},
				{label = 5, type = oil.idl.ulong  , name = "ulong_value"},
				{label = 4, type = oil.idl.float  , name = "float_value"},
				{label = 3, type = oil.idl.double , name = "double_value"},
				{label = 2, type = oil.idl.boolean, name = "boolean_value"},
				{label = 1, type = oil.idl.octet  , name = "octet_value"},
				{label = 0, type = oil.idl.string , name = "string_value"},
			},
		},
		enum = { name = "MyEnum", "MON", "TUE", "WED", "THU", "FRI" },
		sequence = { oil.idl.string },
		array    = { oil.idl.string, length = 5 },
		typedef  = { name = "MyType", oil.idl.string },
		except   = { name = "MyExcept",
			{type = oil.idl.short  , name = "short_value"},
			{type = oil.idl.ushort , name = "ushort_value"},
			{type = oil.idl.long   , name = "long_value"},
			{type = oil.idl.ulong  , name = "ulong_value"},
			{type = oil.idl.float  , name = "float_value"},
			{type = oil.idl.double , name = "double_value"},
			{type = oil.idl.boolean, name = "boolean_value"},
			{type = oil.idl.octet  , name = "octet_value"},
			{type = oil.idl.string , name = "string_value"},
		},
	}
	
	-- Function that creates echo operation for a idltype
	local function new_echo_ops(idltype, def)
		if type(oil.idl[idltype]) == "function"
			then idltype = oil.idl[idltype](def)
			else idltype = oil.idl[idltype]
		end
		return oil.idl.operation{
			parameters = {{ type = idltype, name = "value" }},
			result = idltype,
		},
		oil.idl.operation{
			parameters = {
				{ type = idltype, name = "input" },
				{ type = idltype, name = "output", mode = "PARAM_OUT" },
			},
		},
		oil.idl.operation{
			parameters = {
				{ type = idltype, name = "value", mode = "PARAM_INOUT" },
			},
		}
	end

	-- Create the IDL interface without echo operations
	EchoInterface = oil.idl.interface{
		name = "Echo",
		members = {finished = oil.idl.attribute{ oil.idl.boolean }},
	}

	-- Create echo operations for all supported IDL types
	for idltype, def in pairs(IDLTypes) do
		EchoInterface.members["echo"..idltype],
		EchoInterface.members["in_out"..idltype],
		EchoInterface.members["inout"..idltype]
		= new_echo_ops(idltype, def)
	end
]]

Servers = {
[[
	-- Create interface
]]..CreateInterface..[[

	-- Create servant
	local function echo(self, value) return value end
	
	-- Create object
	local impl = setmetatable({}, {
		__index = function(self, op)
			if
				string.find(op, "^echo") or
				string.find(op, "^in_out") or
				string.find(op, "^inout")
			then
				return echo
			end
		end,
	})

	OiLTestRegister("MarshalTestObject", oil.newobject(impl, EchoInterface))
]],
}

Client = [[
	-- Create interface
]]..CreateInterface..[[

	-- Register interface
	oil.Manager:putiface(EchoInterface)

	-- Create proxy
	local proxy = oil.narrow(OiLTestLookUp("MarshalTestObject"))

	local TypeTests = {
		null = { actual = {n=1} },
		void = { actual = {n=1} },
		boolean = {
			actual = {true, false, n=3},
			expected = {true, false, false},
		},
		octet = {
			actual   = {0, 128.9, 128, 255},
			expected = {0, 128  , 128, 255},
		},
		short = {
			actual   = {-32768, -256, -256.9, 0, 256.9, 256, 32767},
			expected = {-32768, -256, -256  , 0, 256  , 256, 32767},
		},
		ushort = {
			actual   = {0, 32768.9, 32768, 65535},
			expected = {0, 32768  , 32768, 65535},
		},
		long = {
			actual   = {-2147483648, -32768, -32768.9, 0, 32768.9, 32768, 2147483647},
			expected = {-2147483648, -32768, -32768  , 0, 32768  , 32768, 2147483647},
		},
		ulong = {
			actual   = {0, 2147483648.9, 2147483648, 4294967295},
			expected = {0, 2147483648  , 2147483648, 4294967295},
		},
		float  = { actual = {0} },
		double = { actual = {-32768.32768, -256.256, 0, 256.256, 32767.32767} },
		char   = { actual = {"a", "\t", "\n", "\0"} },
		string = { actual = { "a", "\t", "\n", "\0",
			"abc\0\1\2\3\t\n",
			"very very very very very very very long string",
		}},
	}
	
	local ActualValues = {}
	ActualValues.struct = {
		short_value   = 32767,
		ushort_value  = 65535,
		long_value    = 2147483647,
		ulong_value   = 4294967295,
		float_value   = 0,
		double_value  = 32767.32767,
		boolean_value = true,
		octet_value   = 255,
		string_value  = "very very very very very very very long string",
	}
	ActualValues.except   = ActualValues.struct
	ActualValues.sequence = IDLTypes.enum
	ActualValues.array    = IDLTypes.enum
	ActualValues.typedef  = ActualValues.struct.string_value
	
	local function equals(value, other, history)
		if not history then history = {} end
		local success = false
		if value == other then
			success = true
		elseif type(value) == type(other) then
			if history[value] == other then
				success = true
			elseif not history[value] and type(value) == "table" then
				history[value] = other
				for key, field in pairs(value) do
					local otherfield = other[key]
					if otherfield == nil then
						for otherkey, otherfield in pairs(value) do
							if
								equals(key, otherkey, history) and
								equals(field, otherfield, history)
							then
								success = true
								break
							end
						end
					else
						success = equals(field, otherfield, history)
						if not success then break end
					end
				end
			end
		end
		return success
	end
	
	for _, opname in pairs{"echo", "inout", "in_out"} do

		-- Automatic Tests
		for idltype, testcase in pairs(TypeTests) do
			local operation = proxy[opname..idltype]
			for index, actual in ipairs(testcase.actual) do
				local expected = testcase.expected and testcase.expected[index] or actual
				Assert:assertEquals(expected, operation(proxy, actual),
					"marshal of "..idltype.." failed for operation "..opname)
			end
		end
	
		-- Custom Tests
		for idltype, actual in pairs(ActualValues) do
			local operation = proxy[opname..idltype]
			Assert:assertTrue( equals(actual, operation(proxy, actual)),
				"marshal of "..idltype.." failed for operation "..opname)
		end
	
		-- Enum Test
		local operation = proxy[opname.."enum"]
		for _, actual in ipairs(ActualValues.sequence) do
			Assert:assertEquals(actual, operation(proxy, actual),
				"marshal of enum failed")
		end
		
		-- Union Test
		local function assertUnion(actual, switch, value, field)
			Assert:assertEquals( switch, actual._switch, "marshal of union switch failed")
			Assert:assertEquals( value , actual._value , "marshal of union value failed")
			Assert:assertEquals( field , actual._field , "identificaton of union field failed")
			Assert:assertEquals( value , actual[field] , "shortcut for union field failed")
		end
		local operation = proxy[opname.."union"]
		local values = ActualValues.struct
		local actual
		for _, option in ipairs(IDLTypes.union.options) do
			local switch = option.label
			local field  = option.name
			local value  = values[field]
			
			actual = {
				_switch = switch,
				_value  = value,
			}
			assertUnion(operation(proxy, actual), switch, value, field)

			actual = {
				_field = field,
				_value = value,
			}
			assertUnion(operation(proxy, actual), switch, value, field)

			values._switch = switch
			assertUnion(operation(proxy, values), switch, value, field)
			values._switch = nil
			
			values._field  = field
			assertUnion(operation(proxy, values), switch, value, field)
			values._field  = nil

			actual = { [field] = value }
			assertUnion(operation(proxy, actual), switch, value, field)
		end

	end
]]
