------------------------------------------------------------------------------
-- Project: Serialization of Lua Values                                     --
-- Release: 2.0 alpha                                                       --
-- Title  : Basic Serialization Functionality                               --
-- Author : Renato Maia <maia@inf.puc-rio.br>                               --
-- Date   : 21/12/2004 11:05                                                --
------------------------------------------------------------------------------

local getmetatable = getmetatable
local setmetatable = setmetatable
local loadstring   = loadstring
local tostring     = tostring
local getfenv      = getfenv
local setfenv      = setfenv
local assert       = assert
local unpack       = unpack
local rawset       = rawset
local pairs        = pairs
local error        = error
local type         = type
local next         = next

local string = require "string"
local table  = require "table"
local debug  = require "debug"

module "serialization"

------------------------------------------------------------------------------
local Serializer
------------------------------------------------------------------------------
local Done
local function serializevalue(value)
	if not Done[value] then
		local meta = getmetatable(value)
		if meta and meta.__serialize
			then Done[value] = meta.__serialize(value)
			else return Serializer[type(value)](value)
		end
	end
	return Done[value]
end
function serialize(...)
	local maincall
	if not Done then
		Done = {}
		maincall = true
	end

	for i = 1, table.getn(arg) do
		local value = arg[i]
		arg[i] = serializevalue(value)
	end

	if maincall then Done = nil end

	return table.concat(arg, ",")
end
------------------------------------------------------------------------------
local function serializestring(str)
	return string.format("%q", str)
end
------------------------------------------------------------------------------
local function serializetable(tab)
	if tab == getfenv(0) then return "getvalue('_G')" end

	local id = tostring(tab)

	-- define serialized value for future serlizalizations of this same table
	Done[tab] = table.concat{"loadtable('", id, "')"}

	-- serialize table contents
	local value = {"loadtable('", id, "',{"}
	for key, val in pairs(tab) do
		table.insert(value, "[")
		table.insert(value, serializevalue(key))
		table.insert(value, "]=")
		table.insert(value, serializevalue(val))
		table.insert(value, ",")
	end
	table.insert(value, "})")

	-- serialize table's metatable
	local meta = getmetatable(tab)
	if meta then
		value = {"setmetatable(", table.concat(value), ",", serializevalue(meta), ")"}
	end

	return table.concat(value)
end
------------------------------------------------------------------------------
local function dumpedbytes(func)
	local str = string.dump(func)
	local bytes = {}
	for i = 1, string.len(str) do
		table.insert(bytes, string.byte(str, i))
	end
	return table.concat(bytes, ",")
end
local function getupvalues(func)
	local upvalues = {}
	local up = 1
	while true do
		local name, value = debug.getupvalue(func, up)
		if name == nil then break end
		rawset(upvalues, name, value)
		up = up + 1
	end
	return upvalues
end
local function serializefunction(func)
	local id = tostring(func)
	
	-- define serialized value for future serlizalizations of this same function
	Done[func] = table.concat{"getvalue('", id, "')"}

	-- serialize function code
	local value = {"loadfunction('", id, "',buildstring(unpack{", dumpedbytes(func), "}))"}

	-- serialize function upvalues
	local upvalues = getupvalues(func)
	if next(upvalues) then
		value = {"setupvalues(", table.concat(value), ",", serializetable(upvalues), ")"}
	end

	-- serialize function environment
	local environment = getfenv(func)
	value = {"setenvironment(", table.concat(value), ",", serializetable(environment), ")"}

	return table.concat(value)
end
------------------------------------------------------------------------------
Serializer = {
	['nil'] = tostring,
	['boolean'] = tostring,
	['number'] = tostring,
	['string'] = serializestring,
	['function'] = serializefunction,
	['table'] = serializetable,
}
setmetatable(Serializer, {__index = function(self, type)
	error("unable to serialize '" .. type .. "'")
end})
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
local LoadEnvironment
local Loaded
function load(serialized)
	Loaded = {_G = getfenv(0)}
	local loader = assert(loadstring("return "..serialized, "serialized values"))
	setfenv(loader, LoadEnvironment)
	return loader()
end
------------------------------------------------------------------------------
local function registervalue(id, value)
	Loaded[id] = value
	return value
end
local function getvalue(id)
	return Loaded[id]
end
------------------------------------------------------------------------------
local function setupvalues(func, upvalues)
	local up = 1
	while debug.setupvalue(func, up, upvalues[debug.getupvalue(func, up)]) do
		up = up + 1
	end
	return func
end
------------------------------------------------------------------------------
local function setenvironment(func, environment)
	setfenv(func, environment)
	return func
end
------------------------------------------------------------------------------
local function loadfunction(id, code)
	return registervalue(id, loadstring(code))	
end
------------------------------------------------------------------------------
local function copytable(source, destiny)
	if source then
		if not destiny then destiny = {} end
		for field, value in pairs(source) do
			destiny[field] = value
		end
	end
	return destiny
end
local function loadtable(id, contents)
	if contents then
		local tab = getvalue(id)
		if tab
			then return copytable(contents, tab)
			else return registervalue(id, contents)
		end
	else
		return getvalue(id) or registervalue(id, {})
	end
end
------------------------------------------------------------------------------
LoadEnvironment = {
	unpack = unpack,
	setmetatable = setmetatable,
	getvalue = getvalue,
	buildstring = string.char,
	setupvalues = setupvalues,
	setenvironment = setenvironment,
	loadfunction = loadfunction,
	loadtable = loadtable,
}
